#ifndef MOC_VIDEO_QT_H
#define MOC_VIDEO_QT_H
 
#include <QWidget>
#include <QtWidgets>
#ifdef MPXPLAY_LINK_QTMEDIA
#include <QtMultimediaWidgets/QVideoWidget>
#endif
#include "moc_config.h"
#ifdef MPXPLAY_LINK_QTMEDIA
#include <QtMultimedia/QMediaPlayer>
#endif
#ifdef MPXPLAY_LINK_QTAV
#include <QtAV/QtAV>
#include <QtAVWidgets/QtAVWidgets>
//#define MPXPLAY_QTAV_VIDEORENDER 1
#endif
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
//#include "evr/evrcustompresenter.h"
#ifdef __cplusplus
extern "C" {
#endif
#include <libavcodec/avcodec.h>
#include <libavutil/imgutils.h>
#include <libswscale/swscale.h>
#ifdef __cplusplus
}
#endif
#endif

QT_BEGIN_NAMESPACE
class MainWindow;
class DispQtVideoEventFilter;
QT_END_NAMESPACE

//#define DISPQT_FFMPEG_TEST_SPEED 1 // test execution duration of different parts (decoding and rendering time)
//#define USE_SWS_IMAGE_SCALING 1  // currently doesn't work (keeped for EVR)
//#define DISPQT_USE_OPENGL_WIDGET 1 // not finished

#define DISPQT_VIDEO_OPEN_TIMEOUT  5000
#define DISPQT_VIDEO_MUTEX_TIMEOUT DISPQT_MUTEX_TIMEOUT  // 5000

#define DISPQT_VIDEO_CURSORHIDE_TIMEOUT 3000

#define DISPQT_VIDEOWIDGET_SINGLECLICK_TIMEOUTMS 300
#define DISPQT_VIDEOWIDGET_DOUBLECLICK_TIMEOUTMS 300

#define DISPQT_VIDEOWIDGET_CMENUHIDE_TIMEOUTMS 100

#define DISPQT_FFMPEG_WORKERTHREAD_REFRESH_MS 10

class DispQtVideoSurfaceContextMenu : public QMenu
{
Q_OBJECT
public:
	DispQtVideoSurfaceContextMenu(MainWindow *main_window = 0, DispQtVideoEventFilter *parent = 0, QWidget *filtered = 0);
	~DispQtVideoSurfaceContextMenu();
protected:
	void mouseReleaseEvent(QMouseEvent *mevent);

private Q_SLOTS:
	void play_start_or_pause(void);
	void actfunc_video_ar_select(QAction *selected_ar_action);
	void actfunc_video_crop_select(QAction *selected_crop_action);
	void actfunc_video_zoom_select(QAction *selected_zoom_action);
	void actfunc_video_program_select(QAction *selected_ar_action);
	void actfunc_video_stream_select(QAction *selected_ar_action);

private:
	MainWindow *main_window;
	DispQtVideoEventFilter *parent_object;
	QWidget *filtered_widget;
	struct mpxplay_programinfo_t *program_list;
	struct mpxplay_streaminfo_t *stream_list;

	QMenu *videocontmenu_sub_audio;
	QMenu *videocontmenu_sub_video;
	QMenu *videocontmenu_subvideo_sub_aspectratio;
	QMenu *videocontmenu_subvideo_sub_crop;
	QMenu *videocontmenu_subvideo_sub_zoom;
	QMenu *videocontmenu_sub_programs;
	QMenu *videocontmenu_sub_streams;

	QAction *menuaction_filename, *action_play_start, *action_play_stop;
	QAction *action_play_skipback, *action_play_skipforward, *action_view_videofs, *action_file_exit;

	QAction *actions_video_ar_select[DISPQT_VIDEOFULLSCREENARTYPE_MAX];
	QAction *actions_video_crop_select[DISPQT_VIDEOSCRCROPTYPE_MAX];
	QAction *actions_video_zoom_select[DISPQT_VIDEOFULLSCREENZOOMTYPE_MAX];
	QAction **actions_program_select;
	QAction **actions_stream_select[MPXPLAY_STREAMTYPEINDEX_NUM];
};

class DispQtVideoEventFilter : public QObject
{
Q_OBJECT
public:
	DispQtVideoEventFilter(MainWindow *main_window = 0, QWidget *parent = 0, QWidget *filtered = 0);
	bool get_mousecursor_forbidden_hide(void) { return this->mouse_cursor_forbidden_hide; }
	void mouse_cursor_show_timed(bool timed);
	void mouse_cursor_forbid_hide(bool forbid);
	void video_surface_move(int move_x, int move_y);
	void video_zoom_step(int globalpos_x, int globalpos_y, int direction); // 1: zoomin, -1:zoomout
	void video_callback_call(int globalpos_x, int globalpos_y, unsigned long command, void *arg1, int *arg2);
	DispQtVideoSurfaceContextMenu *video_surface_contextmenu;

private Q_SLOTS:
	void mouse_cursor_hide(void);
	void cmenu_about_hide(void);
	void mouse_singleclick_timeouthandler(void);

protected:
	bool eventFilter(QObject *watched, QEvent *event);

private:
	bool is_video_fullscreen(void);
	MainWindow *main_window;
	QWidget *parent_widget, *filtered_widget;
	Qt::MouseButton mouse_pressed_buttons;
	int mouse_last_globalpos_x, mouse_last_globalpos_y;
	bool mouse_cursor_forbidden_hide, singleclick_event, singleclick_fullsrc_start, fullsrc_curr, is_cursor_overrided;
	QTimer *timer_singleclick, timer_doubleclick, *timer_mouse_hide, timer_cmenu_hide;
	QMutex mutex_video;
};

#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG

#define INFFMPEG_ASS_DECODED_MAX_LINES    8
#define INFFMPEG_ASS_DECODED_FLAG_ITALIC (1 << 0)

#define INFFMPEG_VIDEO_RENDER_MIN_INTERVAL_MS  15  // we don't display more than 66 frames / second (drop the output frame)
#define INFFMPEG_VIDEO_INITERROR_LIMIT 100  // ca. 1000ms wait with 10ms sleep

typedef struct ffmpegvideo_subtitle_info_s
{
	int last_render_x, last_render_y, last_size_x, last_size_y;
	unsigned int nb_lines;
	char *line_texts[INFFMPEG_ASS_DECODED_MAX_LINES];
	unsigned int line_flags[INFFMPEG_ASS_DECODED_MAX_LINES];
}ffmpegvideo_subtitle_info_s;

extern void video_ffmpeg_subtitleinfo_clear(ffmpegvideo_subtitle_info_s *sub_lines);
extern void video_ffmpeg_subtitleinfo_copy(ffmpegvideo_subtitle_info_s *dest_lines, ffmpegvideo_subtitle_info_s *sub_lines);

#define DISPQT_FFMPGVIDEO_MAX_QUEUED_VIDEOFRAME  5
#define DISPQT_FFMPGVIDEO_VIDEO_MAX_FRAME_DIFF_US   30*1000000  // max half minute difference between video frames

#define DISPQT_FFMPGVIDEOCALLBACKFLAG_EXITEVENT           (1 << 0)  // stop thread event
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_THEXITDONE          (1 << 1)  // done
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_FIRSTFRAME          (1 << 4)  // first displayed frame sign
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO        (1 << 5)  // primary video file
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_DISPLAYONEFRAME     (1 << 6)  // display one video frame (in pause)
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP     (1 << 7)  // ignore timestamp sync (after seeking) to display a frame
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_STILLPICT           (1 << 8)  // still picture (aspect ratio and multiply picture frame workaround)

typedef struct ffmpegvideo_callback_list_s{
	unsigned int flags;
	mpxplay_infile_callback_t *infile_callback;
	void *infile_passdata;
	int videoctx_init_error_count;
	mpxp_int64_t subtitle_begin_timestamp;
	mpxp_int64_t subtitle_end_timestamp;
	mpxp_int64_t video_last_timestamp;
	mpxp_int64_t video_last_render_time;
	mpxplay_packetlist_t *video_pktlist_elem;
	mpxplay_packetlist_t *subtitle_pktlist_elem;
	mpxplay_packetlist_t *videoframe_pktlist_elem;
	ffmpegvideo_subtitle_info_s *subtitle_infos;
	mpxp_thread_id_type videodecoder_thread_id;
	mpxplay_packetqueue_t videoframe_queue;
	struct ffmpegvideo_callback_list_s *next;
}ffmpegvideo_callback_list_s;

class FFMpegVideoDecoderWorker : public QObject
{
Q_OBJECT
public:
	FFMpegVideoDecoderWorker(MainWindow *main_window = 0, QWidget *parent = 0);
	~FFMpegVideoDecoderWorker();
	bool videothread_ffmpeg_config_callback(unsigned int cfgcmd,mpxplay_infile_callback_t *cbk_func, void *passdata);
	int videothread_ffmpeg_callback_call(int globalpos_x, int globalpos_y, unsigned long command, void *arg1, int *arg2);

Q_SIGNALS:
	void signal_ffmpeg_frame_display(int video_index, unsigned int flags, AVFrame *videoout_frame, ffmpegvideo_subtitle_info_s *subtitle_infos);

public Q_SLOTS:
	void TimedEvent(void);

private:
	int cb_queue_put_entry(mpxplay_infile_callback_t *cbk_func, void *passdata, bool insert_to_begin);
	int cb_queue_del_entry(mpxplay_infile_callback_t *cbk_func, void *passdata);
	void cb_queue_elem_clear(struct ffmpegvideo_callback_list_s *cbelem);
	void cb_queue_clear(void);

	ffmpegvideo_callback_list_s *cblist_first, *cblist_last;
	int nb_callbacks;
#ifdef DISPQT_FFMPEG_TEST_SPEED
	mpxp_int64_t timedevent_prev_start_time, timedevent_prev_renderout_time;
#endif
	bool last_is_video_open, is_paused;
	QMutex mutex_decoderthread;
};

#if (DISPQT_USE_OPENGL_WIDGET == 2)
class FFMpegVideoWidget : public QOpenGLWidget
#else
class FFMpegVideoWidget : public QWidget
#endif
{
Q_OBJECT
public:
	FFMpegVideoWidget(MainWindow *main_window = 0, QWidget *parent = 0);
	~FFMpegVideoWidget();
	bool videowidget_ffmpeg_config_callback(unsigned int cfgcmd, mpxplay_infile_callback_t *cbk_func, void *passdata);
	int videowidget_ffmpeg_callback_call(int globalpos_x, int globalpos_y, unsigned long command, void *arg1, int *arg2);

	int brightness() const;
	int contrast() const;
	int hue() const;
	int saturation() const;

Q_SIGNALS:
	void signal_ffmpeg_swctx_resize(int window_width, int window_height, bool do_repaint);
	void signal_ffmpeg_swctx_move(int move_x, int move_y);
	void signal_ffmpeg_swctx_zoom(int globalpos_x, int globalpos_y, int direction);

public Q_SLOTS:
	void videowidget_ffmpeg_swctx_resize(int window_width, int window_height, bool do_repaint);
	void videowidget_ffmpeg_swctx_move(int move_x, int move_y);
	void videowidget_ffmpeg_swctx_zoom(int globalpos_x, int globalpos_y, int direction);
	void videowidget_ffmpeg_frame_display(int video_index, unsigned int flags, AVFrame *videoout_frame, ffmpegvideo_subtitle_info_s *subtitle_infos);
	void setFullScreen(bool fullScreen);
	void setAspectRatioMode(Qt::AspectRatioMode mode);
	void setBrightness(int brightness);
	void setContrast(int contrast);
	void setHue(int hue);
	void setSaturation(int saturation);

protected:
    void resizeEvent(QResizeEvent *e);
    void paintEvent(QPaintEvent *event);

private:
    void videowidget_ffmpeg_render_subtitle(QPainter *painter, ffmpegvideo_subtitle_info_s *subtitle_infos);
	int  videowidget_ffmpeg_videoctx_init(void);
	void videowidget_ffmpeg_videoctx_close(void);
	void videowidget_ffmpeg_videoctx_clear(void);
	bool videowidget_ffmpeg_swsctx_crop_auto_detect(uint8_t *video_data);
	bool videowidget_ffmpeg_swsctx_calculate(int window_width, int window_height);
	int  videowidget_ffmpeg_swsctx_init(int width, int height);
	void videowidget_ffmpeg_swsctx_close(void);
	void videowidget_ffmpeg_swsctx_clear(void);

	MainWindow *main_window;
	QWidget *parent_widget;
	AVFrame *ffmpeg_sws_frame;
	bool ffmpeg_swsframe_hasdata, ffmpeg_swsframe_clearpic;
	uint8_t *sws_buffer;
	struct SwsContext *sws_ctx;
	unsigned long last_video_render_type, last_video_crop_type, ffmpeg_swsframe_cropreset;
	int last_video_zoom_min_val, last_video_zoom_ti_val;
	int sws_src_width, sws_src_height;
	int sws_srccrop_x, sws_srccrop_y, sws_srccrop_width, sws_srccrop_height;
	int sws_crop_x, sws_crop_y, sws_crop_w, sws_crop_h;
	int video_zoom_globalpos_x, video_zoom_globalpos_y;
	int sws_dst_width, sws_dst_height, sws_dst_pos_x, sws_dst_pos_y;
	int sws_autoxcrop_framecounter;
	enum AVPixelFormat sws_src_pix_fmt;
	AVRational video_aspect_ratio;
	QMutex mutex_video;
	FFMpegVideoDecoderWorker *video_decoder_worker;
	ffmpegvideo_subtitle_info_s *sub_title_infos;
#if (DISPQT_USE_OPENGL_WIDGET == 1)
	QOpenGLContext *video_opengl_context;
#endif
//	Scheduler *video_sample_scheduler_evr;
//	EVRCustomPresenter *video_renderer_evr;
	QThread video_decoder_thread;
	QTimer video_decoder_timer;
	QFont font_subtitle;
};
#endif

class DispQtVideoPlayer : public QWidget
{
Q_OBJECT
public:
enum {VideoCtrlValue_FullScreen = 0, VideoCtrlValue_Height, VideoCtrlValue_Brightness, VideoCtrlValue_Contrast, VideoCtrlValue_Hue, VideoCtrlValue_Saturation, VideoCtrlValue_Speed};
	DispQtVideoPlayer(MainWindow *main_window = 0, QWidget *parent = 0);
	~DispQtVideoPlayer();
	void video_mediaplayer_open(bool lock_qmedia);
	void video_mediaplayer_close(void);
	void set_focus(void) { if(this->video_widget_context) this->video_widget_context->setFocus(Qt::MouseFocusReason); }
	void video_request_open(void *handle_ptr, char *utf8_filename);
	bool video_wait_for_open(void);
	void video_request_close(void *handle_ptr);
	bool video_wait_for_close(void);
	bool video_is_fullscreen(void) { return ((this->video_widget_context)? this->video_widget_context->isFullScreen() : false); }
	qint64 video_tell_ms(void);
	qint64 video_duration_ms(void);
	QWidget *video_get_widgetptr(void) { return this->video_widget_context; }
    int  video_get_value(unsigned int ctrl);
	void video_mouse_cursor_show(bool timed);
	void video_mouse_cursor_forbid_hide(bool forbid);
	void video_widget_init(void);
	void video_widget_control(unsigned long mainwin_control);
	//QString video_media_name(void) { return this->media_name; }
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	int video_ffmpeg_configcallback_create(unsigned int cfgcmd, mpxplay_infile_callback_t *cbk_func, void *passdata);
#endif

Q_SIGNALS:
	void signal_video_open(void);
	void signal_video_close(void);
	void signal_video_play(bool startflag);
	void signal_video_pause(void);
	void signal_video_switch_play(void);
	void signal_video_seek_ms(qint64 pos_ms);
	void signal_video_set_fullscreen(bool fs);
	void signal_video_switch_fullscreen(void);
	void signal_video_apply_fullscreen(bool fs); // 2nd step of switch_fullscreen
	void signal_video_check_windows(void);
	void signal_video_apply_windows(void);       // 3rd step of switch_fullscreen
	void signal_video_event_stop(void);
	void signal_video_set_speed(int speed1000);
	void signal_audio_set_volume(int vol);
	void signal_video_set_value(unsigned int ctrl, int val);
	void signal_video_surface_refresh(void);
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	void signal_ffmpeg_config_callback(unsigned int cfgcmd, mpxplay_infile_callback_t *cbk_func, void *passdata);
#endif

private Q_SLOTS:
	void video_open(void);
	void video_close(void);
	void video_play(bool startflag);
	void video_pause(void);
	void video_switch_play(void) { if(this->is_playing) this->video_pause(); else this->video_play(false); }
	void video_stop(void *handle_ptr, bool close_media);
	void video_seek_ms(qint64 pos_ms);
	void video_set_fullscreen(bool fs);
	void video_switch_fullscreen(void);
	void video_apply_fullscreen(bool fs);
	void video_check_windows(void);
	void video_apply_windows(void);
	void video_handle_event_stop(void);
	void video_set_speed(int speed1000);
	void audio_set_volume(int vol);
	void video_set_value(unsigned int ctrl, int val);
	void video_surface_refresh(void);
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	void video_ffmpeg_configcallback_wrapper(unsigned int cfgcmd, mpxplay_infile_callback_t *cbk_func, void *passdata);
	void video_ffmpeg_config_callback(unsigned int cfgcmd, mpxplay_infile_callback_t *cbk_func, void *passdata);
#endif

#ifdef MPXPLAY_LINK_QTMEDIA
	void status_changed(QMediaPlayer::MediaStatus status);
	void display_error_message(QMediaPlayer::Error);
#endif

private:
#ifdef MPXPLAY_LINK_QTMEDIA
	void video_cursor_select(QMediaPlayer::MediaStatus status);
#endif

	MainWindow *main_window;
	QWidget *parent_widget;
	QHBoxLayout *video_layout;
	QString media_name;
#ifdef MPXPLAY_LINK_QTAV
	QtAV::AVPlayer *qtav_media_player;
#ifdef MPXPLAY_QTAV_VIDEORENDER
	QtAV::VideoRenderer *qtav_video_output;
#else
	QtAV::VideoOutput *qtav_video_output;
#endif
#endif
#ifdef MPXPLAY_LINK_QTMEDIA
	QMediaPlayer *qt_media_player;
	QUrl media_url;
	QMediaContent media_content;
#endif
	DispQtVideoEventFilter *video_event_filter;
	QWidget *video_widget_context;
	Qt::WindowFlags video_wdg_nonfullscreen_flags;
	bool is_video_show, is_video_open, is_wait_for_open, is_playing;
	bool is_wait_for_close, is_video_closed;
	void *ptr_handle_curr, *ptr_handle_new, *ptr_handle_close; // for close identification
	QMutex mutex_player, mutex_qmedia;
};

#endif
